(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import rel_utils.
Require Import list_utils.

Require Import bar_t.

Set Implicit Arguments.

Section fan_t.

  (* That proof of the FAN theorem is derived from Daniel Fridlender's paper *)

  Notation mono := ((fun X (P : list X -> Prop) => forall x l, P l -> P (x::l)) _).

  Variables (X : Type) (P : list X -> Prop) (HP : mono P).
  
  Let P_mono l m : P m -> P (l++m).
  Proof.
    induction l; simpl; auto.
  Qed.
  
  Let V u := fun lw => Forall (fun v => P (v++u)) (list_fan lw).

  Let V_mono u : mono (V u).
  Proof.
    unfold V; intros w lw H.
    apply list_fan_Forall.
    rewrite Forall_forall; intros x _.
    revert H; apply Forall_impl.
    intro; apply HP.
  Qed.

  Let bar_V u : bar_t P u -> bar_t (V u) nil.
  Proof.    
    induction 1 as [ u Hu | u Hu IHu ] using bar_t_rect.
    
    apply in_bar_t0; unfold V.
    simpl. 
    constructor; simpl; auto.
    
    apply in_bar_t1; intros w.
    induction w as [ | a w IHw ].
    
    apply in_bar_t0.
    unfold V; simpl.
    constructor.

    specialize (IHu a).
    apply bar_t_lift1 in IHw.
    unfold V in IHw.
    generalize (@V_mono (a::u)); intros HV1.
    assert (mono (fun lw => V u (lw++w::nil))) as HV2.
      intros x ll; simpl; apply V_mono.
    generalize (bar_t_intersection HV1 HV2 IHu IHw); intros HV3; clear HV1 HV2.
    apply bar_t_lift2.
    revert HV3; apply bar_t_inc.

    intros x; unfold V.
    intros [ H1 H2 ].
    rewrite Forall_perm with (1 := list_fan_middle_cons x a w nil).
    apply Forall_app; split; auto.
    rewrite list_fan_sg_right.
    rewrite Forall_map.
    revert H1; apply Forall_impl.
    intro; rewrite app_ass; auto.
  Qed.

  Theorem fan_t_on_list : bar_t P nil 
                       -> bar_t (fun ll => Forall P (list_fan ll)) nil.
  Proof.
    intros H.
    generalize (bar_V H).
    apply bar_t_inc.
    intros lw; unfold V.
    apply Forall_impl.
    intro; rewrite <- app_nil_end; auto.
  Qed.

End fan_t.



