(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.
Require Import sublist.

Require Import good_base.
Require Import bar_t.

Set Implicit Arguments.

Section Bar_Good.

  Variables (A : Type) (R : A -> A -> Prop).

  Infix "<<" := R (at level 70).

  (* every infinite sequence contains a Good prefix *)

  Theorem bar_t_good_nil_seq : bar_t (good R) nil -> forall f, { m | good R (pfx_rev f m) }.
  Proof.
    apply bar_t_inv.
  Qed.

  Definition bar_t_good_nil_all : bar_t (good R) nil -> forall ll, bar_t (good R) ll. 
  Proof.
    apply bar_t_nil_all.
    intros ? ?; apply in_good_1.
  Qed.

  Fact bar_t_good_app ll mm : bar_t (good R) mm -> bar_t (good R) (ll++mm).
  Proof.
    apply bar_t_app.
    intros; apply in_good_1; auto.
  Qed.

  Theorem sublist_bar_t_good ll mm : ll <sl mm -> bar_t (good R) ll -> bar_t (good R) mm.
  Proof.
    intros H1 H2; revert H2 mm H1.
    induction 1 as [ ll Hll | ll Hll IH] using bar_t_rect; intros mm Hmm.
    apply in_bar_t0; revert Hll; apply sublist_good; auto.
    apply in_bar_t1; intros a.
    apply (IH a); constructor 2; auto.
  Qed.

  Fact bar_t_good_refl a : bar_t (good R) nil -> a << a.
  Proof.
    intros H.
    apply ctxt_nil.
    revert H.
    generalize (@nil A).
    induction 1 as [ ll Hll | ll Hll IH ] using bar_t_rect.
    left; auto.
    specialize (IH a).
    apply ctxt_cons in IH; tauto.
  Qed.

  Fact bar_t_lift_rel ll a : bar_t (good (R rlift a)) ll -> bar_t (good R) (ll++a::nil).
  Proof.
    induction 1 using bar_t_rect.
    apply in_bar_t0; apply good_lift_rel; auto.
    apply in_bar_t1; auto.
  Qed.

  Section bar_t_lift_rel.
  
    Let lemma2 a ll : good R (ll++a::nil) -> bar_t (good (R rlift a)) ll.
    Proof.
      intros H.
      apply in_bar_t1; intros u.
      apply in_bar_t0.
      apply good_app_inv in H.
      
      destruct H as [ H | [ H | (x & b & H1 & H2 & H3) ] ].
      
      constructor 2.
      revert H; apply good_inc.
      intros ? ?; red; tauto.
      
      apply good_sg_inv in H; destruct H.
      
      destruct H2 as [ H2 | [] ]; subst b.
      constructor 1 with x; auto.
      red; auto.
    Qed.
  
    Let lemma3 mm : bar_t (good R) mm -> forall a ll, mm = ll++a::nil -> bar_t (good (R rlift a)) ll.
    Proof.
      induction 1 as [ mm Hmm | mm Hmm IH ] using bar_t_rect; intros a ll ?; subst.

      apply lemma2; auto.
      
      apply in_bar_t1; intros b.
      apply IH with b; auto.
    Qed.

    Fact bar_t_lift_rel_rev a ll : bar_t (good R) (ll++a::nil) -> bar_t (good (R rlift a)) ll.
    Proof.
      intros ?; apply lemma3 with (2 := eq_refl); auto.
    Qed.

  End bar_t_lift_rel. 

End Bar_Good.
