(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import tacs.

Require Import af.
Require Import afs.

Require Import af_t.
Require Import afs_t.

Require Import tree.
Require Import tree_embeddings.

Require Import kruskal_ktree_af.
Require Import kruskal_ktree_af_t.

Set Implicit Arguments.

Section Kruskal_Theorems.

  Variable (X : Type) (R : X -> X -> Prop).

  (* The logical (or non-informative or prop-bounded) version *)

  Theorem kruskal_af : af R -> af (emb_tree_homeo R).
  Proof.
    intros HR.
    apply af_afs_True_eq in HR.
    generalize (@kruskal_wfptree_afs _ 0 (fun _ _ => True) (fun _ => R)); intros K; spec_all K; auto.
    
    assert (H : forall t, wfptree (fun _ (_ : X) => True) t).
      induction t as [ x ll IH ];
      rewrite wfptree_fix; auto.
      
    apply af_afs_True_eq.
    revert K.
    apply afs_mono.
    intros; auto.
    intros ? ?;  apply emb_tree_homeo_upto_eq; auto.
  Qed.
  
  Theorem kruskal_af_t : af_t R -> af_t (emb_tree_homeo R).
  Proof.
    intros HR.
    apply af_t_afs_t_True in HR.
    generalize (@kruskal_wfptree _ 0 (fun _ _ => True) (fun _ => R)); intros K; spec_all K; auto.
    
    assert (H : forall t, wfptree (fun _ (_ : X) => True) t).
      induction t as [ x ll IH ];
      rewrite wfptree_fix; auto.
    
    apply afs_t_True_af_t.
    revert K.
    apply afs_t_relmap with (f := @eq _); auto.
    intros y; exists y; split; auto.
    intros; subst; auto.
    apply emb_tree_homeo_upto_eq; auto.
  Qed.

End Kruskal_Theorems.

Check kruskal_af.
Print Assumptions kruskal_af.

Check kruskal_af_t.
Print Assumptions kruskal_af_t.

(* Extraction "kruskal.ml" kruskal_af_t. *)