(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import ChoiceFacts.

Require Import list_utils. (* for pfx_rev *)

Section Infinite_recursion.

  Variable (X : Type) (P : list X -> Prop).

  Hypothesis HPd : forall ll, P ll \/ ~ P ll.
  Hypothesis HP0 : P nil.
  Hypothesis HPS : forall ll, P ll -> exists x, P (x::ll).

  Axiom choice : FunctionalDependentChoice_on (list X).

  Let R mm ll := ~ P mm \/ ll <> nil /\ mm = tail ll /\ P ll.
  
  Let myhd (ll : list X) : ll <> nil -> { x | exists mm, ll = x::mm }.
  Proof.
    destruct ll as [ | x ll ].
    intros H; exfalso; apply H; auto. 
    intros _; exists x, ll; auto.
  Qed.
  
  Let R_prop : forall mm, exists ll, R mm ll.
  Proof.
    intros mm.
    destruct (HPd mm) as [ H1 | H1 ].
    apply HPS in H1.
    destruct H1 as (x & H1).
    exists (x::mm); right; repeat split; auto; discriminate.
    exists nil; left; auto.
  Qed.

  Theorem functional_list_dep_choice : exists f, forall n, P (pfx_rev f n).
  Proof.
    destruct (choice R R_prop nil) as (f & H2 & H3).
    assert (forall n, P (f n)) as H1.
      induction n as [ | n IHn ]; simpl; auto.
      rewrite H2; auto.
      destruct (H3 n) as [ | (? & ? & ?) ]; tauto.
    assert (forall n, f (S n) <> nil /\ f n = tail (f (S n))) as H4.
      intros n;
      destruct (H3 n) as [ H | (? & ? & ?) ]; auto.
      exfalso; contradict H; auto.
    set (g n := proj1_sig (myhd _ (proj1 (H4 n)))).
    assert (forall n, f (S n) = (g n)::(f n)) as Hf.
      intros n.
      destruct (proj2_sig (myhd _ (proj1 (H4 n)))) as ( mm & H6 ).
      generalize (proj2 (H4 n)); rewrite H6; simpl.
      intros H; subst; reflexivity.
    assert (forall n, f n = pfx_rev g n) as Hg.
      induction n; simpl; auto.
      rewrite <- IHn; auto.

    exists g.
    intros n.
    rewrite <- Hg; auto.
  Qed.
   
End Infinite_recursion.
