(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Permutation.

Require Omega.

Set Implicit Arguments.

(* Various Permutation results *)

Infix "~p" := (@Permutation _) (at level 80).

Section Permutation_tools.

  Variable X : Type.
  
  Implicit Types (l : list X).
  
  Theorem Permutation_In_inv l1 l2: l1 ~p l2 -> forall x, In x l1 -> exists l, exists r, l2 = l++x::r.
  Proof.
    intros H x Hx.
    apply in_split.
    apply Permutation_in with (1 := H).
    auto.
  Qed.

  Theorem Permutation_add_one l1 l2 a r1 r2 : l1++r1 ~p l2++r2 -> l1++a::r1 ~p l2++a::r2.
  Proof.
    intros H.
    apply Permutation_trans with (2 := Permutation_middle _ _ _).
    apply Permutation_sym.
    apply Permutation_trans with (2 := Permutation_middle _ _ _).
    apply Permutation_sym.
    auto.
  Qed.

  Hint Resolve Permutation_add_one.
    
  Theorem Permutation_app_intro l1 l2 k r1 r2 : l1++r1 ~p l2++r2 -> l1++k++r1 ~p l2++k++r2.
  Proof. induction k; auto; simpl;intro; auto. Qed.

  Theorem Permutation_simpl_middle l1 l2 k r1 r2 : l1++k++r1 ~p l2++k++r2 -> l1++r1 ~p l2++r2.
  Proof.
    induction k; auto.
    intros H; simpl in H.
    apply Permutation_app_inv in H.
    auto.
  Qed.
 
End Permutation_tools.
