(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Omega.

Set Implicit Arguments.

(* a small library to reverse lexicographic products *)

Section nat_reverse.

  Variable k : nat.           

  Definition revk i := 
    match le_lt_dec k i with
      | left _  => i
      | right _ => k - S i
    end.
    
  Fact revk_inv i : revk (revk i) = i.
  Proof.
    unfold revk.
    do 2 destruct (le_lt_dec k i); try omega.
    destruct (le_lt_dec k (k - S i)); omega.
  Qed.
  
  Fact revk_eq1 i : i < k -> revk i = k - S i.
  Proof.
    unfold revk.
    destruct (le_lt_dec k i); omega.
  Qed.
  
  Fact revk_eq2 i : k <= i -> revk i = i.
  Proof.
    unfold revk.
    destruct (le_lt_dec k i); omega.
  Qed.
  
  Fact revk_prop i : i < k -> revk i < k.
  Proof.
    unfold revk.
    destruct (le_lt_dec k i); omega.
  Qed.
  
  Definition rev U (s : nat -> U) i := s (revk i).     
  
  Fact rev_inv U s i : rev (@rev U s) i = s i.
  Proof.
    unfold rev; rewrite revk_inv; auto.
  Qed.

End nat_reverse.
